#!/usr/bin/Rscript --vanilla
##
## Plot the changes in disease burden for each SA3.
##
##
## Usage:
##
##     ./plot-sa3-deltas.R input.ssv output.pdf
##


SCRIPT <- 'plot-sa3-deltas.R'


plot <- function(data_file, plot_file) {
    df <- read.table(data_file, header = TRUE, as.is = TRUE)
    df$SA3 <- factor(df$SA3)

    labels <- c('abs_all' = 'ABS (all modes)',
                'abs_private' = 'ABS (private)',
                'abs_public' = 'ABS (public)',
                'sygic' = 'Sygic')
    df$modality <- labels[df$file]

    p <- ggplot(mapping = aes(x = SA3, y = pcnt_of_cases - pcnt_of_popn,
                              colour = frac_self)) +
        geom_hline(yintercept = 0) +
        geom_point() +
        xlab(NULL) +
        ylab('Change in disease burden') +
        scale_colour_distiller(expression("\u03b4"[i]^H),
                               palette = 'Spectral',
                               limits = c(0, 1), breaks = c(0, 0.5, 1),
                               guide = guide_colourbar(direction = 'horizontal',
                                                       title.vjust = 1)) +
        facet_wrap(~ modality, scales = 'free_y') +
        theme_light() +
        theme(axis.text.x = element_text(angle = 90, hjust = 1, vjust = 0.5),
              legend.position = 'top',
              legend.title = element_text(size = rel(1.25)),
              legend.text = element_text(size = rel(1.25)),
              strip.background = element_rect(size = 0, fill = 'transparent'),
              strip.text = element_text(colour = 'black', size = rel(1.5)),
              strip.text.x = element_text(margin = margin(b = 5)),
              strip.text.y = element_text(margin = margin(l = 10)),
              panel.spacing.x = unit(2, 'lines'),
              panel.spacing.y = unit(1, 'lines'),
              axis.title.y = element_text(size = rel(1.5)),
              plot.title = element_text(hjust = 0.5))

    p1 <- p %+% df[df$frac_cbd == 0.33, ]
    p2 <- p %+% df[df$frac_cbd == 0.67, ]

    p1a <- p1 +
        geom_point(data = df,
                   mapping = aes(x = SA3, y = median - pcnt_of_popn),
                   colour = 'black', shape = 4)

    p2a <- p2 +
        geom_point(data = df,
                   mapping = aes(x = SA3, y = median - pcnt_of_popn),
                   colour = 'black', shape = 4)

    cat ('Writing plot to', plot_file, '...\n')
    CairoPDF(plot_file, bg = 'transparent', w = 10, h = 8)
    print(p1)
    print(p2)
    print(p1a)
    print(p2a)
    invisible(dev.off())
}


load_libs <- function() {
    ## Ensure that the required libraries are available.
    reqd_libs <- c('ggplot2', 'Cairo')
    failed <- c()
    for (lib in reqd_libs) {
        if (! require(lib, character.only = TRUE, quietly = TRUE)) {
            failed <- c(failed, lib)
        }
    }
    if (length(failed) > 0) {
        cat('\nCould not load:', paste(failed, collapse = ', '), '\n\n')
        quit(status = 10)
    }
}


main <- function(args = NULL) {
    if (is.null(args) || length(args) != 2) {
        cat('\nUsage:', SCRIPT, 'input.ssv', 'output.pdf\n\n')
        return(invisible(2))
    }

    load_libs()
    plot(args[1], args[2])

    invisible(0)
}


##
## Only call main() if this script is being run from the command-line.
##
if (! interactive()) {
    file.arg <- grep('^--file=', commandArgs(FALSE), value = TRUE)
    file.path <- substring(file.arg, 8)
    file.regex <- paste0('^(|.*/)', SCRIPT)
    if (grepl(file.regex, file.path)) {
        quit(status = main(args = commandArgs(TRUE)))
    }
}
